# -*- coding: utf-8 -*-

"""MySQL 源实例相关的操作

作者: 蒋乐兴|neeky@live.com
时间: 2023-04
"""

import time
import logging
from pathlib import Path
from dbma.core import messages
from dbma.bil.fun import fname
from dbma.components.mysql.commons import default_pkg
from dbma.components.mysql.install import install_mysql
from dbma.components.mysql.commons import make_mysql_writable


def install_mysql_source(
    port: int = 3306, pkg: Path = default_pkg, innodb_buffer_pool_size: str = "128M"
):
    """安装 MySQL Source 结点

    Parameters
    ----------
    port : int, optional
        MySQL 的端口号, by default 3306
    pkg : Path, optional
        安装包的全路径, by default default_pkg
    innodb_buffer_pool_size : str, optional
        innodb buffer pool 的大小, by default "128M"
    """
    # 安装 MySQL 实例
    logging.info(messages.FUN_STARTS.format(fname()))
    install_mysql(port, pkg, innodb_buffer_pool_size=innodb_buffer_pool_size)

    # 把只读打开
    time.sleep(7)
    make_mysql_writable(port)

    # 返回
    logging.info(messages.FUN_ENDS.format(fname()))



# -*- coding: utf-8 -*-

"""安装 orchestrator
"""

import re
import tarfile
import logging
from pathlib import Path
from dbma.bil.fun import fname
from dbma.core import messages
from dbma.components.orchestrator.exceptions import (
    OrchHasBeenInstalledException,
    OrchPkgNotExistsException,
)


def assert_orch_not_installed():
    """断言 orchestrator 有没有安装，如果配置文件
    /usr/local/orchestrator/orchestrator.conf.json 存在说明安装过了

    Returns:
    --------
    None

    Exceptions:
    -----------
    OrchHasBeenInstalledException
    """
    logging.info(messages.FUN_STARTS.format(fname()))

    orch_config_file = Path("/usr/local/orchestrator/orchestrator.conf.json")
    if orch_config_file.exists():
        logging.warn(messages.FUN_ENDS.format(fname()))
        raise OrchHasBeenInstalledException("orchestrator has been installed .")

    logging.info(messages.FUN_ENDS.format(fname()))


def get_orch_version(pkg: Path):
    """根据 orch 安装包的全路径返回 orch 的版本号

    Parameters:
    -----------
    pkg: Path

    Returns:
    --------
    str
    """
    logging.info(messages.FUN_STARTS.format(fname()))

    if pkg is None:
        raise ValueError("pkg arg is None on call {} function .".format(fname()))

    pkg_name = pkg.name
    p = re.compile(r"orchestrator-(?P<version>\d{1,2}.\d{1,2}.\d{1,2})-.*\.tar.gz")

    logging.info(messages.FUN_ENDS.format(fname()))
    return p.match(pkg_name).group("version")


def decompression_pkg(pkg: Path):
    """解压 orchestrator 的安装包

    Parameters:
    -----------
    pkg:Path
        安装的的全路径

    Exceptions:
    -----------
    ValueError
    OrchPkgNotExistsException
    -----------
    """
    logging.info(messages.FUN_STARTS.format(fname()))
    if pkg is None:
        raise ValueError("pkg arg is None on call decompression_pkg function .")

    # 到这里说明 pkg 至少不为 None
    if not pkg.exists():
        raise OrchPkgNotExistsException(messages.FILE_NOT_EXISTS.format(str(pkg)))

    # flag_file 用于确认有没有解压过
    flag_file = Path("/usr/local/orchestrator-3.2.6/.dbm-agent-decompression.txt")

    # 真正的执行解压操作
    with tarfile.open(pkg) as tar_pkg:
        tar_pkg.extractall("/usr/local/")

    # 解压完成之后写入标记文件 basedir/.dbm-agent-decompression.txt
    with open("", "w") as f:
        f.write("dbm-agent")

    # 解压完成
    logging.info(messages.FUN_ENDS.format(fname()))


def install_orch(pkg: Path):
    """安装 orchestrator 到 /usr/local/

    Parameters:
    -----------
    pkg: Path
        安装包的全路径
    """
    # 断定 orch 没有安装过
    try:
        assert_orch_not_installed()
    except OrchHasBeenInstalledException as err:
        raise err

    # 解压
    # 渲染配置文件
    # 渲染 systemd 文件
    # 启动



# -*- coding: utf-8 -*-

"""Resdis systemd 相关的配置"""

import logging
from pathlib import Path
from jinja2 import Template
from datetime import datetime
from dataclasses import dataclass, asdict

from dbma.bil.cmdexecutor import exe_shell_cmd
from dbma.bil.fun import fname
from dbma.core import messages
from dbma.components.redis.exceptions import RedisConfigTemplateFileNotExistsException


@dataclass
class RedisSystemdConfig(object):
    port: int = 6379
    basedir: str = ""
    now: str = datetime.now().isoformat()

    def render_config(self):
        """沉浸 Redis 的 Systemd 配置文件

        Exceptions:
        -----------
        RedisConfigTemplateFileNotExistsException
        """
        logging.info(messages.FUN_STARTS.format(fname()))
        import dbma

        redis_systemd_config_template = (
            Path(dbma.__file__).parent / "static/cnfs/redisd.service.jinja"
        )
        # 检查模板文件是否存在
        if not redis_systemd_config_template.exists():
            msg = "redis systemd config file '' not exists .".format(
                redis_systemd_config_template
            )
            logging.error(msg)
            logging.error(messages.FUN_ENDS.format(fname()))
            raise RedisConfigTemplateFileNotExistsException(msg)

        # 取出配置文件并渲染
        content = ""
        with open(redis_systemd_config_template) as template_object:
            content = template_object.read()

        template = Template(content)
        logging.info(messages.FUN_ENDS.format(fname()))
        # 把 Path 转为 str
        if isinstance(self.basedir, Path):
            self.basedir = str(self.basedir)
        return template.render(asdict(self))

    def generate_systemd_config(self):
        """生成 systemd 配置文件"""
        logging.info(messages.FUN_STARTS.format(fname()))
        systemd_config_file = Path(
            "/usr/lib/systemd/system/"
        ) / "redisd-{}.service".format(self.port)
        # 保存到文件
        with open(systemd_config_file, "w") as systemd_object:
            systemd_object.write(self.render_config())
        # 生效
        cmd = "systemctl daemon-reload"
        logging.info(cmd)
        exe_shell_cmd(cmd)
        logging.info(messages.FUN_ENDS.format(fname()))



# -*- coding: utf-8 -*-

"""MySQL 安装的相关接口
"""

import re
import logging
from aiohttp import web
from pathlib import Path
from dbma.bil.fun import fname
from dbma.core import messages
from dbma.bil.sudos import sudo
from dbma.core.router import routes
from dbma.core.configs import dbm_agent_config
from dbma.core.threads.backends import threads
from dbma.core.views.response import ResponseEntity
from dbma.components.mysql.commons import default_pkg
from dbma.components.mysql.install import uninstall_mysql
from dbma.components.mysql.instance import is_instance_exists
from dbma.components.mysql.views.handlers import clone_local_data_task_handler
from dbma.components.mysql.views.handlers import install_mysql_source_task_handler
from dbma.components.mysql.views.handlers import install_mysql_replica_task_handler


@routes.view("/apis/mysqls/defaults")
class MySQLDefaultsView(web.View):
    """
    MySQL 的全局默认配置
    """

    async def get(self):
        """返回 MySQL 的全局默认配置"""
        logging.info(messages.VIEW_FUN_STARTS.format(self.request.url))
        resposne = ResponseEntity(
            message="",
            error="",
            data={
                "mysql-datadir-parent": dbm_agent_config.mysql_datadir_parent,
                "mysql-default-version": dbm_agent_config.mysql_default_version,
                "mysql-binlogdir-parent": dbm_agent_config.mysql_binlogdir_parent,
            },
        )
        logging.info(messages.VIEW_FUN_ENDS.format(self.request.url))
        return web.json_response(resposne.to_dict(), status=200)


@routes.view("/apis/mysqls/install")
class MySQLInstallView(web.View):
    """
    MySQL 安装逻辑

    详细步骤:
    1. 检查 post 参数
    2. 如果参数中带了 task-id 就以任务的方式，放后台执行 & 返回结果告诉调用方，任务已经提交
    3. 如果参数中没有 task-id 就同步以同步的方式安装 MySQL & 返回结果告诉调用方，安装失败还是成功
    """

    port: int = None
    ibps: str = None
    pkg_name: str = None
    pkg: str = None
    source: str = None
    role: str = None
    task_id: str = None
    response: ResponseEntity = None

    async def parser_post_args(self):
        """处理 post 参数"""
        logging.info(messages.FUN_STARTS.format(fname()))
        data = await self.request.json()

        if "role" not in data:
            self.response.message = "instance 'role' arg miss, you can choise one of ['master', 'slave', 'source', 'replica']"
            return
        self.role = data["role"]

        if "port" not in data:
            self.response.message = messages.ARG_NOT_IN_POST_DICT.format("port")
            logging.warn(messages.FUN_ENDS.format(fname()))
            return
        self.port = int(data["port"])

        if "ibps" not in data:
            self.response.message = messages.ARG_NOT_IN_POST_DICT.format("ibps")
            logging.warn(messages.FUN_ENDS.format(fname()))
            return
        self.ibps = data["ibps"]

        # 如果没有指定 pkg-name 那么就使用默认值
        if "pkg-name" not in data:
            self.pkg_name = default_pkg.name
        self.pkg_name = data["pkg-name"]
        self.pkg = Path("/usr/local/dbm-agent/pkgs/") / self.pkg_name

        # source 可以没有，如果有那 source 一定要格式正确
        if "source" not in data:
            self.source = None
        else:
            self.source = data["source"]
            p = re.compile("\d{1,3}.\d{1,3}.\d{1,3}.\d{1,3}:\d{1,4}")
            if not p.match(self.source):
                self.response.message = "source arg format error"
                return

        # task-id 可以不传
        if "task-id" not in data:
            self.task_id = None
        else:
            self.task_id = data["task-id"]

        logging.info(messages.FUN_ENDS.format(fname()))

    async def check_args(self):
        """检查参数之间是否有逻辑冲突"""
        logging.info(messages.FUN_STARTS.format(fname()))
        if self.role in ("slave", "replica"):
            if self.source is None:
                self.response.message = (
                    "args role = '{}', you must give 'source' arg ".format(self.role)
                )
                logging.warn(messages.FUN_ENDS.format(fname()))
                return

        logging.info(messages.FUN_ENDS.format(fname()))

    async def install_dispatch(self):
        """根据参数决定怎么安装 MySQL"""
        logging.info(messages.FUN_STARTS.format(fname()))
        # 根据 task_id 是不是 None 来决定接口是同步执行还是异步执行
        if self.task_id is not None:
            # 进入异步处理逻辑
            # ---------------
            if self.role in ("master", "source"):
                # 进入安装单机/主库的处理逻辑
                threads.submit(
                    install_mysql_source_task_handler,
                    port=self.port,
                    ibps=self.ibps,
                    pkg=self.pkg,
                    task_id=self.task_id,
                )
                self.response.message = (
                    "submit install mysql 'master|source' task to backends threads."
                )
            elif self.role in ("slave", "replica"):
                # 进入备机的处理逻辑
                source_ip, source_port = self.source.split(":")
                threads.submit(
                    install_mysql_replica_task_handler,
                    port=self.port,
                    ibps=self.ibps,
                    pkg=self.pkg,
                    source_ip=source_ip,
                    source_port=source_port,
                    task_id=self.task_id,
                )
                self.response.message = (
                    "submit install mysql 'slave|replica' task to backends threads."
                )
            logging.info(messages.VIEW_FUN_ENDS.format(self.request.url))
            # return web.json_response(self.response.to_dict(), status=200)
        else:
            # 进入同步处理逻辑
            # -------------
            if self.role in ("master", "source"):
                # 进入安装单机/主库的处理逻辑
                install_mysql_source_task_handler(
                    port=self.port, ibps=self.ibps, pkg=self.pkg, task_id=self.task_id
                )
                self.response.message = "install mysql 'master|source' complete ."
            elif self.role in ("slave", "replica"):
                # 进入备机的处理逻辑
                source_ip, source_port = self.source.split(":")
                install_mysql_replica_task_handler(
                    port=self.port,
                    ibps=self.ibps,
                    pkg=self.pkg,
                    source_ip=source_ip,
                    source_port=source_port,
                    task_id=self.task_id,
                )
                self.response.message = "install mysql 'slave|replica' complete ."
        logging.info(messages.FUN_ENDS.format(fname()))

    async def post(self):
        logging.info(messages.VIEW_FUN_STARTS.format(self.request.url))
        # 准备返回结果
        self.response = ResponseEntity(message="", error="", data=None)

        # 处理参数
        await self.parser_post_args()
        await self.check_args()
        if self.response.message:
            return web.json_response(self.response.to_dict(), status=500)

        # 打印一下接收到的参数
        logging.info(
            "port = '{}', ibps = '{}', pkg-name = '{}', pkg = '{}', source = '{}', role = '{}' task-id = {} .".format(
                self.port,
                self.ibps,
                self.pkg_name,
                self.pkg,
                self.source,
                self.role,
                self.task_id,
            )
        )

        # 安装
        await self.install_dispatch()
        logging.info(messages.VIEW_FUN_ENDS.format(self.request.url))
        return web.json_response(self.response.to_dict(), status=200)


@routes.view("/apis/mysqls/uninstall")
class MySQLUninstallView(web.View):
    """MySQL 卸载逻辑

    1. 检查 port 参数有没有传递
    2. 检查 port 对应的实例是否存在于当前机器上
    3. 执行删除逻辑
    """

    port: int = None
    response: ResponseEntity = None

    async def parser_post_args(self):
        """处理 post 参数"""
        logging.info(messages.FUN_STARTS.format(fname()))
        data = await self.request.json()
        if "port" not in data:
            self.response.message = messages.ARG_NOT_IN_POST_DICT.format("port")
            return
        self.port = int(data["port"])
        logging.info(messages.FUN_ENDS.format(fname()))

    async def uninstall_dispatch(self):
        """卸载 MySQL"""
        if not is_instance_exists(self.port):
            self.response.message = "instance {} not exists.".format(self.port)
            return

        try:
            with sudo("uninstall mysql '{}' ".format(self.port)):
                uninstall_mysql(self.port)
        except Exception as err:
            self.response.message = "got error on uninstall mysql {} .".format(
                self.port
            )

    async def post(self):
        """卸载 MySQ"""
        logging.info(messages.VIEW_FUN_STARTS.format(self.request.url))
        # 准备返回结果
        self.response = ResponseEntity(message="", error="", data=None)

        # 处理参数
        await self.parser_post_args()
        if self.response.message:
            return web.json_response(self.response.to_dict(), status=500)

        # 卸载
        await self.uninstall_dispatch()

        # 到这里说明正确的完成了
        logging.info(messages.VIEW_FUN_ENDS.format(self.request.url))
        return web.json_response(self.response.to_dict(), status=200)


@routes.view("/apis/mysqls/{port}/exists")
class MySQLInstanceInfoView(web.View):
    port: int = None
    response: ResponseEntity = None

    async def parser_get_args(self):
        """处理 get 参数"""
        logging.info(messages.FUN_STARTS.format(fname()))

        if self.request.match_info.get("port"):
            self.port = int(self.request.match_info.get("port"))
        else:
            self.port = None
            self.response.message = messages.ARG_NOT_IN_POST_DICT.format("port")

        logging.info(messages.FUN_ENDS.format(fname()))

    async def get(self):
        """检查给定端口的 MySQL 数据库实例是否存在"""
        logging.info(messages.VIEW_FUN_STARTS.format(self.request.url))
        # 准备返回结果
        self.response = ResponseEntity(
            message="", error=None, data={"exists": False, "port": None}
        )

        # 处理参数
        await self.parser_get_args()
        if self.response.message:
            return web.json_response(self.response.to_dict(), status=500)

        # 检查实例是否存在并返回结果
        self.response.data["exists"] = is_instance_exists(self.port)
        self.response.data["port"] = self.port

        logging.info(messages.VIEW_FUN_ENDS.format(self.request.url))
        return web.json_response(self.response.to_dict(), status=200)


@routes.view("/apis/mysqls/{port}/backup")
class MySQLBackupView(web.View):
    port: int = None
    backup_type: str = None
    response: ResponseEntity = None

    async def parser_get_args(self):
        """处理 get 参数"""
        logging.info(messages.FUN_STARTS.format(fname()))

        if not self.request.match_info.get("port"):
            self.response.message = messages.ARG_NOT_IN_POST_DICT.format("port")
            return
        self.port = int(self.request.match_info.get("port"))

        logging.info(messages.FUN_ENDS.format(fname()))

    async def parser_post_args(self):
        """处理 post 参数"""
        logging.info(messages.FUN_STARTS.format(fname()))
        data = await self.request.json()
        if "backup-type" not in data:
            self.response.message = messages.ARG_NOT_IN_POST_DICT.format("backup-type")
            return
        self.backup_type = data["backup-type"]
        logging.info(messages.FUN_ENDS.format(fname()))

    async def backup_dispatch(self):
        """"""
        if self.backup_type == "clone":
            threads.submit(
                clone_local_data_task_handler,
                port=self.port,
                backup_type=self.backup_type,
            )
            self.response.message = (
                "submit backup mysql using clone task to backends threads."
            )
            logging.info(messages.VIEW_FUN_ENDS.format(self.request.url))

        raise ValueError("not suported backp type {}".format(self.backup_type))

    async def post(self):
        """处理备份逻辑"""
        logging.info(messages.VIEW_FUN_STARTS.format(self.request.url))
        # 准备返回结果
        self.response = ResponseEntity(message="", error=None, data=None)

        # 处理参数
        await self.parser_get_args()
        await self.parser_post_args()
        if self.response.message:
            return web.json_response(self.response.to_dict(), status=500)

        # 备份
        try:
            await self.backup_dispatch()
            return web.json_response(self.response.to_dict(), status=200)
        except Exception as err:
            self.response.message = str(err)
            self.response.error = err
        logging.warn(messages.VIEW_FUN_ENDS.format(self.request.url))
        return web.json_response(self.response.to_dict(), status=500)



# -*- coding: utf-8 -*-


""" dbm-agent 版本号管理
"""


DBM_AGENT_MAJOR_VERSION = 8
DBM_AGENT_MINOR_VERSION = 33
DBM_AGENT_PATCH_VERSION = 13


# dbm-agent 版本号
DBM_AGENT_VESION = (
    f"{DBM_AGENT_MAJOR_VERSION}.{DBM_AGENT_MINOR_VERSION}.{DBM_AGENT_PATCH_VERSION}"
)
DBM_VERSION_NUMBER = DBM_AGENT_VESION
VERSION = DBM_AGENT_VESION



# -*- coding: utf8 -*-

"""
实现文件系统的相关操作
"""

import os
import tarfile
from pathlib import Path


def is_file_exists(file_path: Path = None):
    """
    判断文件是否存在

    Parameters:
    -----------
    file_path: Path
        文件|目录 的路径

    Return:
    --------
    bool

    Exceptions:
    -----------
    ValueError
        file_path 为 None 时抛出
    TypeError
        file_path 的类型不为 str | Path 时抛出
    """
    if file_path is None:
        raise ValueError("file_path is None , not a valid value .")

    if type(file_path) not in (str, Path):
        raise TypeError("file_path must be a str, or an Path object .")

    # 如果是 str 的话要传成 Path 对象
    if isinstance(file_path, str):
        file_path = Path(file_path)

    return file_path.exists()


def extract_tar_file(tar_file_path, extract_dir):
    """
    解压tar文件

    Parameters:
    ----------
    tar_file_path: str
        等待解压的tar文件路径

    extract_dir: str
        解压到的目标路径

    Return:
    -------
        None
    """
    tar = tarfile.open(tar_file_path)
    tar.extractall(extract_dir)
    tar.close()


def get_tar_file_name(tar_file_path):
    """
    获取tar文件的名称

    Parameters:
    -----------
    tar_file_path: str
        tar 文件的全路径

    Return:
    -------
        str
    """
    tar = tarfile.open(tar_file_path)
    name = tar.getnames()[0]
    tar.close()
    if os.sep in name:
        name, *_ = name.split(os.sep)
    return name


def link(src, dest):
    """
    创建链接

    Parameters:
    -----------
    src: str
        源路径

    dest: str
        目标路径

    Return:
    -------
        None
    """
    os.symlink(src, dest)


def is_line_in_etc_profile(line):
    """
    /etc/profile 中是否有以 line 开头的行存在

    Parameters:
    -----------
    line: str
        行内容

    Return:
    -------
        bool
    """
    with open("/etc/profile", "r") as f:
        for l in f:
            if l.startswith(line):
                return True
    return False


def append_new_line_to_etc_profile(line):
    """
    在 /etc/profile 中添加一行
    Parameters:
    -----------
    line: str
        要写入的新行

    Returns:
    --------
        None
    """
    with open("/etc/profile", "a") as f:
        f.write(line)
        f.write("\n")


def get_file_size(file_path: Path):
    """返回给定文件的大小(字节)

    Parameters:
    -----------
    file_path: Path
        文件路径
    """
    # 如果参数是 str 给它转换成 Path
    if isinstance(file_path, str):
        file_path = Path(file_path)

    res = os.lstat(file_path)
    return res.st_size


def is_file_greater_then(file_path: Path = None, size: int = None):
    """
    检查 file_path 对应文件大小是不是大于 size, 如果是返回 True, 不是返回 False。

    Parameters:
    -----------
    file_path: Path
        文件路径

    size: int
        大小

    Return:
    -------
    bool
    """
    return get_file_size(file_path) > size


def truncate_or_delete_file(file_path: Path = None, chunk_size: int = 16 * 1024 * 1024):
    """
    如果 file_path 的大小大于 chunk_size 就截断 chunk_size 大小的段，不然就删除文件

    Parameters:
    -----------
    file_path: Path
        要执行删除|截断的文件

    chunk_size: int
        单次截断的大小

    """
    if isinstance(file_path, str):
        file_path = Path(file_path)

    # 如果已经不足 chunk_size 大，就删除掉；然后就 truncate 指定大小
    file_size = get_file_size(file_path)
    if file_size < chunk_size:
        os.remove(file_path)
        return 0

    chunk = file_size - chunk_size
    os.truncate(file_path, chunk)
    return chunk


join = os.path.join

readlink = os.readlink

listdir = os.listdir

mkdir = os.mkdir



# -*- coding: utf-8 -*-

"""
Redis 相关的公共函数
"""

import re
from pathlib import Path

from dbma.bil.fs import is_file_exists
from dbma.core.configs import dbm_agent_config

# Redis 的默认端口号
default_redis_port = 6379

# 根据 dbm-agent 的配置文件确认默认情况下使用的 Redis 版本号
default_redis_pkg = Path(
    "/usr/local/dbm-agent/pkgs/redis-7.0.11-linux-glibc-2.34-x86_64.tar.gz"
)

# Redis 安装包的正则表达式
redis_pkg_re_pattern = re.compile(
    r"redis-(?P<redis_version>\d{1,2}.\d{1,2}.\d{1,3})-linux-glibc-(?P<glibc_version>\d.\d{1,2})-x86_64.tar.gz"
)


def port_to_redis_systemd_config_file(port: int = 6378):
    """根据 Port 生成 Redis Systemd 配置文件全路径

    Parameters
    ----------
    port : int, optional
        Redis 的端口号, by default 6378
    """
    return Path("/usr/lib/systemd/system/redisd-{}.service".format(port))


def is_redis_systemd_config_file_exists(port: int = 6378):
    """Redis 的 Systemd 配置文件是否存在

    Parameters
    ----------
    port : int, optional
        Redis 的端口号, by default 6378
    """
    systemd_config_file = port_to_redis_systemd_config_file(port)
    try:
        return is_file_exists(systemd_config_file)
    except TypeError as err:
        return False
    except ValueError as err:
        return False
    except Exception as err:
        return False


def disable_redis_systemd(port: int = 6378):
    """禁用对应实例的 Redis systemd 配置

    Parameters
    ----------
    port : int, optional
        实例的端口号, by default 6378
    """
    pass



# -*- coding: utf-8 -*-

"""MySQL 克隆插件备份逻辑

CLONE LOCAL DATA DIRECTORY [=] 'clone_dir';
"""


import os
import logging
from pathlib import Path
from datetime import datetime
from dbma.core import messages
from dbma.bil.fun import fname
from dbma.bil.osuser import MySQLUser
from dbma.core.configs import dbm_agent_config
from dbma.components.mysql.commons import dbma_mysql_cnx


def backup_format(backup_type: str = None):
    """备份目录最后一级"""
    now = datetime.now().isoformat().replace(":", "-").replace(".", "-")
    if backup_type.lower() == "clone":
        return "{}-backup-{}".format("clone", now)


def clone_local_data(port: int = 3306):
    """数据本机上的 MySQL 实例到本机上的磁盘

    Parameters:
    -----------
    port: int
        MySQL 端口号

    Returns:
    --------
    备份文件保存的目录
    """
    logging.info(messages.FUN_STARTS.format(fname()))
    backup_dir = (
        Path(dbm_agent_config.mysql_backupdir_parent)
        / str(port)
        / backup_format("clone")
    )
    logging.info("backup_di = '{}' ".format(backup_dir))

    if not backup_dir.parent.exists():
        logging.info("dir '{}' not exists .".format(backup_dir.parent))
        user = MySQLUser(port=port)
        os.makedirs(backup_dir.parent)
        user.chown(str(backup_dir.parent))
        logging.info("dir '{}' create complete. ".format(backup_dir.parent))

    with dbma_mysql_cnx(port) as cursor:
        sql = "CLONE LOCAL DATA DIRECTORY '{}';".format(str(backup_dir))
        logging.info("going go execute sql = '{}' ".format(sql))
        cursor.execute(sql)
    return str(backup_dir)



# -*- coding: utf-8 -*-

"""
背景: 
    当我们卸载一个对应的 MySQL 实例之后、它的数据目录、Binlog 目录、配置文件都应该被清理掉。把个比较它之前的数据目录可能是 
    '/database/mysql/data/3306' 当我们执行卸载操作之后并不会真正的删除数据，而是给它 rename 成了 '3306-backup-2023-06-17T14-45-21-949788'
    由于使用的是 rename 所以回退的时候还是比较方便的
    
    但是这个又带来了一个问题，对于类似 '3306-backup-2023-06-17T14-45-21-949788' 的目录我们什么时候清理呢？默认是 3 天之后在后台线程中清理它；这里
    我们就来实现这个功能。
    
    一个线程定期的检查有哪些实例可以清理，生成任务丢到队列里面去；
    另一个线程从队列中取出任务，然后慢慢的删除对应的文件
    
"""

import os
import re
import glob
import time
import shutil
import logging
from pathlib import Path
from datetime import datetime
from collections import deque
from dataclasses import dataclass
from dbma.bil.fun import fname
from dbma.bil.fs import truncate_or_delete_file, get_file_size
from dbma.core import messages
from dbma.core.configs import dbm_agent_config
from dbma.core.threads.backends import threads, keep_threads_running
from dbma.bil.osuser import sudo


clear_tasks = deque(maxlen=1024)

backup_dir_re_pattern = re.compile(
    r"(?P<port>[0-9]{4,4})-backup-(?P<dt>\d{4}-\d{1,2}-\d{1,2}T\d{1,2}-\d{1,2}-\d{1,2}-\d{4,6})"
)


@dataclass
class ClearTask(object):
    """
    清理任务对象
    """

    path: Path

    @property
    def rename_at(self):
        """根据目录名，计算出目录备份的时间点"""
        # 目录的后缀是 “2023-05-29T20-23-32-472128” 这个格式，也就是说它不是一个正确的时间日期格式
        # 如果我们要得到 datetime 对象，还要给它整一下才行
        match = backup_dir_re_pattern.search(str(self.path))
        dt_str = match.group("dt")
        date_str, tm_str = dt_str.split("T")
        hour, minte, seconde, ms = tm_str.split("-")
        datetime_str = "{} {}:{}:{}.{}".format(date_str, hour, minte, seconde, ms)
        return datetime.fromisoformat(datetime_str)

    def is_expired(self, now=None):
        """
        检查 rename 文件是否已经过期了
        """
        match = backup_dir_re_pattern.search(str(self.path))
        # 用传好格式构造 datetime 对象
        if match:
            now = datetime.now() if now is None else now
            delta = now - self.rename_at
            if delta.days >= 3:
                return True
        # 没有匹配到正则、或是没有超过 3 天
        return False


def scan_data_dir_gen_task():
    """只扫描 binlog 目录和 data 目录；至于 backup 目录这个是交由备份子系统完成

    Returns:
    --------
    [ClearTask,ClearTask ...]
    """
    logging.info(messages.FUN_STARTS.format(fname()))

    result = []
    target_dirs = [
        dbm_agent_config.mysql_datadir_parent,
        dbm_agent_config.mysql_binlogdir_parent,
    ]
    for path in target_dirs:
        # 处理一下路径的格式，让它可以满足 glob.glob 的要求
        if path.endswith("/"):
            target = path + "*"
        else:
            target = path + "/*"

        logging.info("scan dir '{}' .".format(target))

        # 逐个比较找到 ${port}-backup-xxx 格式的备份目录
        for instance_path in glob.glob(target):
            if backup_dir_re_pattern.search(instance_path):
                logging.info("find '{}' .".format(instance_path))
                # 构造 ClearTask 对象
                task = ClearTask(instance_path)
                if task.is_expired():
                    result.append(task)
                else:
                    logging.info(
                        "instance '{}' backup not expired .".format(instance_path)
                    )

    logging.info(messages.FUN_ENDS.format(fname()))
    return result


def clear_instance(task: ClearTask = None):
    """
    根据 ClearTask 中指定的目录进行清理动作
    """
    logging.info(messages.FUN_STARTS.format(fname()))
    logging.info(
        "task.path = '{}'  is_expire = '{}' ".format(task.path, task.is_expired())
    )
    files = []
    dirs = []

    for item in glob.glob("{}/*".format(task.path)):
        path = Path(item)
        if path.is_dir():
            dirs.append(path)
        else:
            files.append(path)
    logging.info("files = {}, dirs = {}".format(len(files), len(dirs)))

    # 先清理文件
    for path in files:
        logging.info("deal-with file '{}' ".format(path))
        # 准备清理
        while True:
            # 如果文件比较大，那么就一直 truncate 到 0 为止
            chunck = truncate_or_delete_file(path, 16 * 1024 * 1024)
            time.sleep(1)
            if chunck == 0:
                # chunck == 0 说明文件已经执行 remove 清理掉了
                logging.info("file '{}' removed ".format(path))
                break
            else:
                logging.info("file '{}' truncated ".format(path))

    # 清理子目录
    for sub in dirs:
        clear_instance(ClearTask(sub))

    # 清理当前目录
    dirs = []
    for item in glob.glob("{}/*".format(task.path)):
        path = Path(item)
        if path.is_dir():
            dirs.append(path)

    if len(dirs) == 0:
        logging.info(
            "sub directorys not exists, rm current directory '{}' ".format(task.path)
        )
        shutil.rmtree(task.path)

    logging.info(messages.FUN_ENDS.format(fname()))


def pub_clear_task_thread_fun():
    """
    生成后台清理任务的线程函数
    """
    global keep_threads_running
    while keep_threads_running:
        logging.info(messages.FUN_STARTS.format(fname()))
        with sudo():
            tasks = scan_data_dir_gen_task()
            for task in tasks:
                clear_tasks.append(task)

        logging.info(messages.FUN_STARTS.format(fname()))

        # 一小时扫一次目录，生成清理任务
        time.sleep(3600)


def sub_clear_task_thread_fun():
    """
    从队列里取出任务并执行清理
    """
    global keep_threads_running
    while keep_threads_running:
        try:
            logging.info(messages.FUN_STARTS.format(fname()))
            try:
                task = clear_tasks.pop()
            except IndexError as err:
                logging.info("task deque is empty .")
                # 对于队列中没有任务的情况下线程休息 30 分钟
                time.sleep(1800)
                continue

            with sudo():
                clear_instance(task)
            # 清理完成一个实例之后要休息 1 分钟
            time.sleep(60)
            logging.info(messages.FUN_STARTS.format(fname()))
        except Exception as err:
            logging.error(err)


def start_clear_tasks():
    threads.submit(pub_clear_task_thread_fun)
    time.sleep(5)
    threads.submit(sub_clear_task_thread_fun)



# -*- coding: utf-8 -*-

"""Resdis 相关的异常"""

from dbma.core.exception import (
    FileExistsException,
    FileNotExistsException,
    DirectoryExistsException,
)


class RedisConfigTemplateFileNotExistsException(FileNotExistsException):
    """Redis 配置文件模板不存在"""

    pass


class RedisPkgFileNotExistsException(FileNotExistsException):
    """Redis 的安装包文件不存在"""

    pass


class RedisDatabaseDirectoryExistsException(DirectoryExistsException):
    """Redis 的数据目录已经存在"""

    pass



# -*- coding: utf-8 -*-

"""dbm-agent 收到操作请求(task)之后，如果请求中没有带 task_id ，说明要求同步执行；如果有带 task_id 那么我们要把 task 放到后台线程中执行。

"""

import logging
from pathlib import Path
from dbma.bil.sudos import sudo
from dbma.components.mysql.backups.cloneplugin import clone_local_data
from dbma.components.mysql.source import install_mysql_source
from dbma.components.mysql.replica import install_mysql_replica


def update_task_state_callback(
    task_id: int = None, state: str = None, message: str = None
):
    """调用 dbm-center 的接口更新回调任务的状态"""
    logging.info("starts update task state callback .")
    # TODO 调用更新接口ends
    logging.info("task_id = {}".format(task_id))
    logging.info("ends update task state callback .")


def install_mysql_source_task_handler(
    port: int = 3306, ibps: str = "128M", pkg: Path = None, task_id: int = None
):
    """让安装 MySQL 的逻辑放后台执行

    Parameters:
    -----------

    port: int
        MySQL 端口号

    ibps: str
        innodb_buffer_pool_size 的大小， et 128M, 2G, 8G ...

    pkg: str
        MySQL 安装包全路径

    Return:
    -------
    None
    """
    logging.info("starts install mysql 'master|source' task handler .")
    try:
        # 提升后台线程的权限到 root
        with sudo("install mysql task handler"):
            install_mysql_source(port=port, innodb_buffer_pool_size=ibps, pkg=pkg)
        logging.info("install mysql complete")

        # 是否更新任务信息到 dbm-center
        if not task_id is None:
            logging.warn(
                "install mysql task handler's callback function is None, skip callback"
            )
        else:
            update_task_state_callback(task_id, 200, "install mysql complete")

    except Exception as err:
        logging.error("install mysql task handler got error ")
    logging.info("ends install mysql task handler .")


def install_mysql_replica_task_handler(
    port: int = 3306,
    ibps: str = "128M",
    pkg: Path = None,
    source_ip: str = None,
    source_port: str = None,
    task_id: int = None,
):
    """让安装 MySQL 备机的逻辑放后台执行

    Parameters:
    -----------

    port: int
        MySQL 端口号

    ibps: str
        innodb_buffer_pool_size 的大小， et 128M, 2G, 8G ...

    pkg: str
        MySQL 安装包全路径

    Return:
    -------
    None
    """
    logging.info("starts install mysql 'slave|replica' task handler .")
    try:
        # 提升后台线程的权限到 root
        with sudo("install mysql task handler"):
            # install_mysql(port=port, innodb_buffer_pool_size=ibps, pkg=pkg)
            install_mysql_replica(
                port=port,
                pkg=pkg,
                innodb_buffer_pool_size=ibps,
                source_ip=source_ip,
                source_port=source_port,
            )
        logging.info("install mysql 'slave|replica' complete")

        # 是否更新任务信息到 dbm-center
        if not task_id is None:
            logging.warn(
                "install mysql 'slave|replica' task handler's callback function is None, skip callback"
            )
        else:
            update_task_state_callback(
                task_id, 200, "install mysql 'slave|replica' complete"
            )

    except Exception as err:
        logging.error("install mysql 'slave|replica' task handler got error ")
    logging.info("ends install mysql 'slave|replica' task handler .")


def clone_local_data_task_handler(
    port: int = 3306, backup_type: str = None, task_id: int = None
):
    """ """
    logging.info("starts clone local data task handler .")
    try:
        with sudo("starts clone local data task handler"):
            clone_local_data(port)
    except Exception as err:
        logging.error("starts clone local data task handler got error {}".format(err))

    # 是否更新任务信息到 dbm-center
    if not task_id is None:
        logging.warn(
            "starts clone local data task handler's callback function is None, skip callback"
        )
    else:
        update_task_state_callback(
            task_id, 200, "starts clone local data task handler complete"
        )

    logging.info("ends clone local data task handler .")



# -*- coding: utf-8 -*-

"""Redis 环境的安装"""


import os
import tarfile
import logging
from pathlib import Path
from dbma.core import messages
from dbma.bil.fun import fname
from dbma.bil.osuser import RedisUser
from dbma.components.redis.exceptions import (
    RedisDatabaseDirectoryExistsException,
    RedisPkgFileNotExistsException,
)
from dbma.components.redis.commons import (
    default_redis_pkg,
    redis_pkg_re_pattern,
    default_redis_port,
)
from dbma.bil.cmdexecutor import exe_shell_cmd
from dbma.components.redis.config import RedisConfig, RedisReplicaConfig
from dbma.components.redis.systemd import RedisSystemdConfig


def create_redis_user(port: int = 6379):
    """创建 redis 用户"""
    logging.info(messages.FUN_STARTS.format(fname()))

    # 根据端口号创建用户
    redis_user = RedisUser(port)
    redis_user.create()

    logging.info(messages.FUN_ENDS.format(fname()))


def chown_database_dir_to_redis_user(port: int = 6379):
    """更新数据目录属组到 Redis 用户

    Parameters
    ----------
    port : int, optional
        Redis 端口号, by default 6379
    """
    logging.info(messages.FUN_STARTS.format(fname()))

    redis_user = RedisUser(port)
    database_dir = Path("/database/redis/{}".format(port))
    logging.info("going to chown .")
    redis_user.chown(str(database_dir), recursive=True)
    logging.info("chown down.")

    logging.info(messages.FUN_ENDS.format(fname()))


def create_redis_database_dir(port: int = 6379):
    """创建 Redis 数据的工作目录
    Paramters:
    ----------
    port: int
        Redis 端口号

    Exceptions:
    -----------
    RedisDatabaseDirectoryExistsException
    """
    logging.info(messages.FUN_STARTS.format(fname()))

    database_dir = Path("/database/redis/{}".format(port))
    if database_dir.exists():
        message = "redis database directory '{}' exists .".format(database_dir)
        raise RedisDatabaseDirectoryExistsException(message)
    # 创建数据目录
    os.makedirs(database_dir)
    chown_database_dir_to_redis_user(port)

    logging.info(messages.FUN_ENDS.format(fname()))


def pkg_to_redis_basedir(pkg: Path = default_redis_pkg):
    """根据 pkg 的内容计算出对应的 basedir 的内容

    Parameters:
    ----------
    pkg : Path, optional
        Redis 的安装包, by default default_redis_pkg

    Returns:
    --------
    Path

    Exceptions:
    -----------
    ValueError: 给定的 pkg 与预先定义的正则对不上
    """
    logging.info(messages.FUN_STARTS.format(fname()))

    # 1. 确认当前 redis 的版本号
    m = redis_pkg_re_pattern.match(pkg.name)
    if m is None:
        logging.error(messages.FUN_ENDS.format(fname()))
        raise ValueError("redis pkg name error pkg-name = '{}'.".format(pkg))
    version = m.group("redis_version")

    # 2. 返回 redis 版本应该对应的 basedir
    logging.info(messages.FUN_ENDS.format(fname()))
    return Path("/usr/local/redis-{}".format(version))


def pkg_to_redis_version(pkg: Path = default_redis_pkg):
    """_summary_

    Parameters
    ----------
    pkg : Path, optional
        Redis 的安装包程序, by default default_redis_pkg
    """
    return redis_pkg_re_pattern.match(pkg.name).group("redis_version")


def decompression_redis_pkg(pkg: Path = default_redis_pkg):
    """解压 Redis 安装到 /usr/local/

    1、检查有没有解压过，如果有的话就路过
    2、检查给定的安装包是否存在、如果不存在就报错
    3、解压
    4、写入标识文件

    Parameters
    ----------
    pkg : Path, optional
        redis 安装包的全路径, by default default_redis_pkg
    """
    logging.info(messages.FUN_STARTS.format(fname()))

    # 1. 检查标识文件是否存在、如果存在说明解压过了，那么现在可以跳过这步解压的过程。
    flag_file = pkg_to_redis_basedir(pkg) / ".dbm-agent-decompression.txt"
    if flag_file.exists():
        logging.info(messages.FUN_ENDS.format(fname()))
        return

    # 2. 检查 pkg 文件是否存在
    if not pkg.exists():
        msg = "redis pkg file '{}' not exists. ".format(str(pkg))
        logging.error(msg)
        raise RedisPkgFileNotExistsException(msg)

    # 3. 解压
    with tarfile.open(pkg) as tf:
        tf.extractall("/usr/local/")

    # 4. 创建标识文件
    with open(flag_file, "w") as flag_object:
        flag_object.write("dbma-agent")

    logging.info(messages.FUN_ENDS.format(fname()))


def start_redis(port: int = 6379):
    """启动 Redis

    Parameters
    ----------
    port : int, optional
        Redis 的端口号, by default 6379
    """
    exe_shell_cmd("systemctl start redisd-{}".format(port))


def stop_redis(port: int = 6379):
    """关闭 Redis

    Parameters
    ----------
    port : int, optional
        Redis 端口号, by default 6379
    """
    exe_shell_cmd("systemctl stop redisd-{}".format(port))


def install_redis(
    port: int = 6379, pkg: Path = default_redis_pkg, redis_config: RedisConfig = None
):
    """安装 Redis 并让其监控到指定端口

    Parameters:
    -----------
    port : int, optional
        Redis 要监听的端口号, by default 6379
    pkg : Path, optional
        Redis 的安装包, by default default_redis_pkg

    Exceptions:
    -----------

    """
    try:
        # 第一步：创建用户
        create_redis_user(port)
        # 第二步：创建数据目录
        create_redis_database_dir(port)
        # 第三步：解压
        decompression_redis_pkg(pkg)
        # 第四步：生成配置
        redis_config.generate_config_file()
        # 第五步：生成 systemd 配置
        redis_systemd_config = RedisSystemdConfig(port, pkg_to_redis_basedir(pkg))
        redis_systemd_config.generate_systemd_config()
        # 第六步：启动 redis
        start_redis(port)
    except Exception as err:
        msg = str(err)
        logging.exception(err)
        raise err


def install_resdis_master(port: int = 6379, pkg: Path = default_redis_pkg):
    """安装 Redis master 结点

    Parameters
    ----------
    port : int, optional
        Redis 端口号, by default 6379
    pkg : Path, optional
        Redis 安装包, by default default_redis_pkg
    """
    logging.info(messages.FUN_STARTS.format(fname()))

    redis_master_config = RedisConfig(port)
    install_redis(port, pkg=pkg, redis_config=redis_master_config)

    logging.info(messages.FUN_ENDS.format(fname()))


def install_redis_replica(
    port: int = 6379, replicaof: str = "127.0.0.1 6379", pkg: Path = default_redis_pkg
):
    """安装 Redis replica 结点

    Parameters
    ----------
    port : int, optional
        Redis 结点的端口号, by default 6379
    master : str, optional
        Redis Master 结点的标识("host port"), by default "127.0.0.1 6379"
    pkg : Path, optional
        Redis 的安装包, by default default_redis_pkg
    """
    logging.info(messages.FUN_STARTS.format(fname()))
    logging.info("redis-port={}, redis-master={}, pkg={}".format(port, replicaof, pkg))

    redis_replica_config = RedisReplicaConfig(port=port, replicaof=replicaof)
    install_redis(port, pkg=pkg, redis_config=redis_replica_config)

    logging.info(messages.FUN_ENDS.format(fname()))



# -*- coding: utf-8 -*-

"""异常树"""

from dbma.core.exception import DBMAgentException


class MySQLTemplateFileNotExistsException(DBMAgentException):
    """MySQL 配置文件模板不存在"""

    pass


class MySQLPkgFileNotExistsException(DBMAgentException):
    """MySQL 安装包文件模板不存在"""

    pass


class InstanceHasBeenInstalledException(DBMAgentException):
    """给定实例已经安装过了"""

    pass


class MySQLSystemdFileNotExists(DBMAgentException):
    """ """

    pass


class MySQLSystemdFileExists(DBMAgentException):
    pass


class MySQLDataDirectoryExists(DBMAgentException):
    """MySQL 数据目录已经存在"""

    pass


class NotSuportMySQLDirectoryType(DBMAgentException):
    """不支持的 MySQL 目录类型"""

    pass
